@@ -1,10 +1,15 @@
 """
-Simple SQLite interface.
+A simple SQLite interface.
 
-SQLite is a relational database management system contained in a C
-programming library. In contrast to many other database management
-systems, SQLite is not a client–server database engine. Rather, it
-is embedded into the end program.
+[SQLite](https://sqlite.org) is a C-language library that implements a small,
+fast, self-contained, high-reliability, full-featured, SQL database engine.
+In contrast to many other database management systems, SQLite is not a
+client–server database engine. Rather, it is embedded into the end program.
+
+[sqlite3](https://docs.python.org/3/library/sqlite3) is a Python interface to
+SQLite.
+
+`sqlyte` provides an opinionated interface on top of `sqlite3`.
 
 """
 
@@ -55,16 +60,21 @@ def from_json(val):
             if key not in dct:
                 return
             item = dct[key]
-            if not item[0]:
-                return
             tz = None
-            if isinstance(item[0], dict):
-                val = item[0]["datetime"]
-                tz = item[0]["timezone"]
+            # XXX print(type(item))
+            if isinstance(item, dict):
+                val = item["datetime"]
+                tz = item["timezone"]
             elif isinstance(item, list):
-                val = item[0]
+                if isinstance(item[0], dict):
+                    val = item[0]["datetime"]
+                    tz = item[0]["timezone"]
+                else:
+                    val = item[0]
             else:
                 val = item
+            if not val:
+                return
             if val[-6] in "-+":
                 val = "".join(val.rpartition(":")[::2])
             try:
@@ -73,7 +83,11 @@ def from_json(val):
                 dt = "?"
             else:
                 if tz:
-                    dt = dt.astimezone(pendulum.timezone(tz))
+                    try:
+                        dt = dt.in_timezone(tz)
+                    except pendulum.tz.zoneinfo.exceptions.InvalidTimezone:
+                        # XXX print("tz conversion failing silently...")  # TODO use log
+                        pass
             dct[key] = [dt]
 
         upgrade_date("published")
@@ -83,6 +97,20 @@ def from_json(val):
     return json.loads(val, object_hook=f)
 
 
+def dttz_to_iso(raw_dttz):
+    """
+    Return the ISO format of given `raw_dttz`.
+
+    `raw_dttz` is a JSON string in the form of {"datetime": ..., "timezone": ...}).
+    """
+    dttz = json.loads(raw_dttz)
+    return (
+        pendulum.parse(dttz["datetime"])
+        .astimezone(pendulum.timezone(dttz["timezone"]))
+        .isoformat()
+    )
+
+
 class Model:
     def __init__(self, name, **schemas):
         self.name = name
@@ -115,6 +143,7 @@ class ModelController:
         self.db = db
 
     def __getattr__(self, attr):
+        # XXX TODO print(self.controllers, self.db)
         return functools.partial(self.controllers[attr], self.db)
 
 
@@ -124,10 +153,26 @@ model = Model
 class JSONEncoder(json.JSONEncoder):
     def default(self, obj):
         if isinstance(obj, (datetime.date, datetime.datetime)):
-            return obj.isoformat()
+            obj = pendulum.instance(obj)
+            return {
+                "datetime": obj.in_timezone("UTC").isoformat(),
+                "timezone": obj.tzinfo.name,
+            }
         return json.JSONEncoder.default(self, obj)
 
 
+# class JSONEncoder(json.JSONEncoder):
+#     def default(self, obj):
+#         if isinstance(obj, uri.URI):
+#             return str(obj)
+#         if isinstance(obj, (datetime.date, datetime.datetime)):
+#             return {
+#                 "datetime": obj.in_tz("utc").isoformat(),
+#                 "timezone": obj.timezone.name,
+#             }
+#         return json.JSONEncoder.default(self, obj)
+
+
 sqlite3.register_converter("JSON", from_json)
 sqlite3.register_adapter(dict, lambda val: JSONEncoder(indent=2).encode(val))
 
@@ -188,7 +233,11 @@ class Database:
 
             setattr(self, command, single_statement_cursor(command))
 
-        conn = sqlite3.connect(path, detect_types=sqlite3.PARSE_DECLTYPES)
+        conn = sqlite3.connect(
+            path, detect_types=sqlite3.PARSE_DECLTYPES, isolation_level=None
+        )
+        conn.execute("PRAGMA journal_mode=WAL")
+        conn.create_function("dttz_to_iso", 1, dttz_to_iso)
         # TODO conn.cursor().execute("PRAGMA foreign_keys = ON;")
 
         # TODO try:
@@ -338,6 +387,7 @@ def db(path, *models) -> Database:
             dbi.insert("_models", name=model.name, version=model.version)
             current_models[model.name] = model.version
             continue
+        # XXX TODO print("V", current_version, model.version)
         if current_version == model.version:
             # TODO check schema in code against schema in db, suggest migration
             continue  # model exists and is up-to-date
